@@ -4,8 +4,12 @@
  */
 
 
+/* XXX trim includes */
 #include <sys/param.h>
 #include <sys/sysctl.h>
+#include <sys/time.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 
 #include <err.h>
 #include <errno.h>
@@ -15,6 +19,7 @@
 #include <string.h>
 #include <unistd.h>
 
+
 #include "BootCache.h"
 
 struct playlist_header {
@@ -519,6 +524,8 @@ BC_print_statistics(char *fname, struct BC_statistics *ss)
 	}
 	if (ss->ss_strategy_blocked > 0)
 		fprintf(fp, "callers blocked           %u\n", ss->ss_strategy_blocked);
+	if (ss->ss_strategy_stolen > 0)
+		fprintf(fp, "stolen page bypasses      %u\n", ss->ss_strategy_stolen);
 	if ((ss->ss_wait_time.tv_sec > 0) || (ss->ss_wait_time.tv_usec > 0)) {
 		msec = (ss->ss_wait_time.tv_sec * 1000) + (ss->ss_wait_time.tv_usec / 1000);
 		if (msec > 0) {
@@ -608,3 +615,35 @@ BC_tag_history(void)
 	}
 	return(0);
 }
+
+/*
+ * Unload the BootCache kext.
+ */
+int
+BC_unload(void)
+{
+	pid_t	child;
+	char	*argv[4];
+	int	result;
+
+	child = fork();
+	switch (child) {
+	case -1:
+		/* fork failed, bail with error */
+		return(errno);
+	case 0:
+		/* we are the child, do our work */
+		argv[0] = BC_KEXTUNLOAD;
+		argv[1] = "-b";
+		argv[2] = BC_BUNDLE_ID;
+		argv[3] = NULL;
+		result = execve(BC_KEXTUNLOAD, argv, NULL);
+		exit((result != 0) ? 1 : 0);
+	default:
+		/* we are the parent, wait for the child */
+		waitpid(child, &result, 0);
+		break;
+	}
+	/* EBUSY is not a good error */
+	return((result != 0) ? EBUSY : 0);
+}
